/*
	SVNService modifications written by Magnus Norddahl.

	Disclaimer: Yes this code is so ugly (that I added, not the original Craig Link stuff)
	that it may burn down your PC and kill your cat! But what can I say.. its C, and all
	I got is the platform SDK here. Bit violations of TCHAR stuff (dont use the _TEXT things
	always, some hardcoded widechar) and some strings should be defined in header file.

*/
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (C) 1993-1997  Microsoft Corporation.  All Rights Reserved.
//
//  MODULE:   service.c
//
//  PURPOSE:  Implements functions required by all services
//            windows.
//
//  FUNCTIONS:
//    main(int argc, char **argv);
//    service_ctrl(DWORD dwCtrlCode);
//    service_main(DWORD dwArgc, LPTSTR *lpszArgv);
//    CmdInstallService(int argc, char **argv);
//    CmdRemoveService();
//    CmdDebugService(int argc, char **argv);
//    ControlHandler ( DWORD dwCtrlType );
//    GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize );
//
//  COMMENTS:
//
//  AUTHOR: Craig Link - Microsoft Developer Support
//

// shut up! 
#pragma warning(disable:4115)
#pragma warning(disable:4127)
#pragma warning(disable:4706)

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>
#include <time.h>
#include <crtdbg.h>

#include "service.h"
#include "log_messages.h"

// internal variables
SERVICE_STATUS          ssStatus;       // current status of the service
SERVICE_STATUS_HANDLE   sshStatusHandle;
DWORD                   dwErr = 0;
BOOL                    bDebug = FALSE;
TCHAR                   szErr[256];

// internal function prototypes
VOID WINAPI service_ctrl(DWORD dwCtrlCode);
VOID WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv);
VOID CmdInstallService(int argc, char **argv);
VOID CmdSetupService(int argc, char **argv);
VOID CmdRemoveService();
VOID CmdDebugService(int argc, char **argv);
BOOL WINAPI ControlHandler ( DWORD dwCtrlType );
LPTSTR GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize );

//
//  FUNCTION: main
//
//  PURPOSE: entrypoint for service
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    main() either performs the command line task, or
//    call StartServiceCtrlDispatcher to register the
//    main service thread.  When the this call returns,
//    the service has stopped, so exit.
//
void _CRTAPI1 main(int argc, char **argv)
{
    SERVICE_TABLE_ENTRY dispatchTable[] =
    {
        { TEXT(SZSERVICENAME), (LPSERVICE_MAIN_FUNCTION)service_main },
        { NULL, NULL }
    };

    if ( (argc > 1) &&
         ((*argv[1] == '-') || (*argv[1] == '/')) )
    {
        if ( _stricmp( "install", argv[1]+1 ) == 0 )
        {
            CmdInstallService(argc-2, argv+2);
        }
        else if ( _stricmp( "remove", argv[1]+1 ) == 0 )
        {
            CmdRemoveService();
        }
        else if ( _stricmp( "setup", argv[1]+1 ) == 0 )
        {
            CmdSetupService(argc-2, argv+2);
        }
        else if ( _stricmp( "debug", argv[1]+1 ) == 0 )
        {
            bDebug = TRUE;
            CmdDebugService(argc, argv);
        }
        else
        {
            goto dispatch;
        }
		_CrtDumpMemoryLeaks();
        exit(0);
    }

    // if it doesn't match any of the above parameters
    // the service control manager may be starting the service
    // so we must call StartServiceCtrlDispatcher
    dispatch:
        // this is just to be friendly
        printf( "%s -install <svnserve params> to install the service\n", SZAPPNAME );
        printf( "%s -remove                    to remove the service\n", SZAPPNAME );
        printf( "%s -setup <svnserve params>   to change command line parameters for svnserve\n", SZAPPNAME );
        printf( "%s -debug                     to run as a console app for debugging\n", SZAPPNAME );
        printf( "\nStartServiceCtrlDispatcher being called.\n" );
        printf( "This may take several seconds.  Please wait.\n" );

        if (!StartServiceCtrlDispatcher(dispatchTable))
            AddToMessageLog(Event_Service, TEXT("StartServiceCtrlDispatcher failed."), EVENTLOG_ERROR_TYPE);
}



//
//  FUNCTION: service_main
//
//  PURPOSE: To perform actual initialization of the service
//
//  PARAMETERS:
//    dwArgc   - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    This routine performs the service initialization and then calls
//    the user defined ServiceStart() routine to perform majority
//    of the work.
//
void WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)
{

    // register our service control handler:
    //
    sshStatusHandle = RegisterServiceCtrlHandler( TEXT(SZSERVICENAME), service_ctrl);

    if (!sshStatusHandle)
        goto cleanup;

    // SERVICE_STATUS members that don't change in example
    //
    ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwServiceSpecificExitCode = 0;


    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;


    ServiceStart( dwArgc, lpszArgv );

cleanup:

    // try to report the stopped status to the service control manager.
    //
    if (sshStatusHandle)
        (VOID)ReportStatusToSCMgr(
                            SERVICE_STOPPED,
                            dwErr,
                            0);

    return;
}



//
//  FUNCTION: service_ctrl
//
//  PURPOSE: This function is called by the SCM whenever
//           ControlService() is called on this service.
//
//  PARAMETERS:
//    dwCtrlCode - type of control requested
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
VOID WINAPI service_ctrl(DWORD dwCtrlCode)
{
    // Handle the requested control code.
    //
    switch(dwCtrlCode)
    {
        // Stop the service.
        //
        // SERVICE_STOP_PENDING should be reported before
        // setting the Stop Event - hServerStopEvent - in
        // ServiceStop().  This avoids a race condition
        // which may result in a 1053 - The Service did not respond...
        // error.
        case SERVICE_CONTROL_STOP:
            ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 0);
            ServiceStop();
            return;

        // Update the service status.
        //
        case SERVICE_CONTROL_INTERROGATE:
            break;

        // invalid control code
        //
        default:
            break;

    }

    ReportStatusToSCMgr(ssStatus.dwCurrentState, NO_ERROR, 0);
}



//
//  FUNCTION: ReportStatusToSCMgr()
//
//  PURPOSE: Sets the current status of the service and
//           reports it to the Service Control Manager
//
//  PARAMETERS:
//    dwCurrentState - the state of the service
//    dwWin32ExitCode - error code to report
//    dwWaitHint - worst case estimate to next checkpoint
//
//  RETURN VALUE:
//    TRUE  - success
//    FALSE - failure
//
//  COMMENTS:
//
BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint)
{
    static DWORD dwCheckPoint = 1;
    BOOL fResult = TRUE;


    if ( !bDebug ) // when debugging we don't report to the SCM
    {
        if (dwCurrentState == SERVICE_START_PENDING)
            ssStatus.dwControlsAccepted = 0;
        else
            ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;

        ssStatus.dwCurrentState = dwCurrentState;
        ssStatus.dwWin32ExitCode = dwWin32ExitCode;
        ssStatus.dwWaitHint = dwWaitHint;

        if ( ( dwCurrentState == SERVICE_RUNNING ) ||
             ( dwCurrentState == SERVICE_STOPPED ) )
            ssStatus.dwCheckPoint = 0;
        else
            ssStatus.dwCheckPoint = dwCheckPoint++;


        // Report the status of the service to the service control manager.
        //
        if (!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus))) {
            AddToMessageLog(Event_Service, TEXT("SetServiceStatus"), EVENTLOG_ERROR_TYPE);
        }
    }
    return fResult;
}


// Debug level defaults at zero:
static int debug_level = 0;
static FILE *debug_file = NULL;


//
//  FUNCTION: AddToMessageLog(LPTSTR lpszMsg)
//
//  PURPOSE: Allows any thread to log an error message
//
//  PARAMETERS:
//    eventID - Event ID.
//    lpszMsg - text for message
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
static char buffer[1024];
VOID AddToMessageLog(int eventID, LPCTSTR lpszMsg, int error_type)
{
    HANDLE  hEventSource;
    LPTSTR  lpszStrings[2];
	time_t ltime;
	char date[128];
	int i, len;

	if (debug_file != NULL)
	{
		time(&ltime);
		strcpy(date, ctime(&ltime));
		date[strlen(date)-1] = 0;

		sprintf(buffer, "%s: %s\n", date, lpszMsg);

		len = strlen(buffer);
		for (i=0; i<len-1; i++)
		{
			if (buffer[i] == '\n') buffer[i] = '\t';
			if (buffer[i] == '\r') buffer[i] = '\n';
		}

		fwrite(buffer, strlen(buffer), 1, debug_file);
		fflush(debug_file);
	}

    if ( !bDebug && eventID != Event_Debug )
    {
		if (debug_file != NULL) return;
        dwErr = GetLastError();

        // Use event logging to log the error.
        //
        hEventSource = RegisterEventSource(NULL, TEXT(SZSERVICENAME));

        lpszStrings[0] = (LPTSTR) lpszMsg;

        if (hEventSource != NULL) {
            ReportEvent(hEventSource, // handle of event source
                (WORD) error_type/*EVENTLOG_ERROR_TYPE*/,  // event type
                0,                    // event category
                eventID,              // event ID
                NULL,                 // current user's SID
                1,                    // strings in lpszStrings
                0,                    // no bytes of raw data
                lpszStrings,          // array of error strings
                NULL);                // no raw data

            (VOID) DeregisterEventSource(hEventSource);
        }
    }
	else
	{
		printf("MessageLog (Event ID %d): %s\r\n", eventID, lpszMsg);
	}
}

//
//  FUNCTION: AddDebug(int debuglevel, LPTSTR lpszMsg)
//
//  PURPOSE: Write debug information if the error level allows it.
//
//  PARAMETERS:
//    debuglevel - level this debug information belong to.
//    lpszMsg - text for message
//
//  RETURN VALUE:
//    none
//
void _AddDebug(int level, LPTSTR lpszMsg)
{
	if (level > debug_level) return;

	AddToMessageLog(Event_Debug, lpszMsg, EVENTLOG_INFORMATION_TYPE);
}
//////////////////////////////////////////////////////////////////////////////

//
//  FUNCTION: SetDebugLevel(int debuglevel)
//
//  PURPOSE: Changes the debug level at which things are written to the event log.
//
//  PARAMETERS:
//    debuglevel - New debug level to use.
//
//  RETURN VALUE:
//    none
//
void SetDebugLevel(int new_level)
{
	debug_level = new_level;
}
//////////////////////////////////////////////////////////////////////////////

//
//  FUNCTION: LogToFile(const char *filename)
//
//  PURPOSE: Changes all logging to be sent to file.
//
//  PARAMETERS:
//    filename - Log filename.
//
//  RETURN VALUE:
//    none
//
void LogToFile(const char *filename)
{
	if (debug_file != NULL) fclose(debug_file);

	debug_file = fopen(filename, "a+t");
	if (debug_file == NULL) debug_file = fopen(filename, "wt");
}
//////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////
//
//  The following code handles service installation and removal
//


//
//  FUNCTION: CmdInstallService()
//
//  PURPOSE: Installs the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdInstallService(int argc, char **argv)
{
    SC_HANDLE   schService;
    SC_HANDLE   schSCManager;

    TCHAR szPath[512];

	HKEY log_key;
	LONG err;

    if ( GetModuleFileName( NULL, szPath, 512 ) == 0 )
    {
        _tprintf(TEXT("Unable to install %s - %s\n"), TEXT(SZSERVICEDISPLAYNAME), GetLastErrorText(szErr, 256));
        return;
    }

    schSCManager = OpenSCManager(
                        NULL,                   // machine (NULL == local)
                        NULL,                   // database (NULL == default)
                        SC_MANAGER_ALL_ACCESS   // access required
                        );
    if ( schSCManager )
    {
        schService = CreateService(
            schSCManager,               // SCManager database
            TEXT(SZSERVICENAME),        // name of service
            TEXT(SZSERVICEDISPLAYNAME), // name to display
            SERVICE_ALL_ACCESS,         // desired access
            SERVICE_WIN32_OWN_PROCESS,  // service type
            SERVICE_DEMAND_START,       // start type
			// Could also be SERVICE_AUTO_START for automatic startup.
            SERVICE_ERROR_NORMAL,       // error control type
            szPath,                     // service's binary
            NULL,                       // no load ordering group
            NULL,                       // no tag identifier
            TEXT(SZDEPENDENCIES),       // dependencies
            NULL,                       // LocalSystem account
            NULL);                      // no password

        if ( schService )
        {
            _tprintf(TEXT("%s installed.\n"), TEXT(SZSERVICEDISPLAYNAME) );
			// To automatically start service ::StartService(schService, 0, NULL);
            CloseServiceHandle(schService);
        }
        else
        {
            _tprintf(TEXT("CreateService failed - %s\n"), GetLastErrorText(szErr, 256));
        }

        CloseServiceHandle(schSCManager);
    }
    else
        _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));


	// Try to register the log message file:

	err = RegCreateKeyEx(
		HKEY_LOCAL_MACHINE,
		TEXT("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\SVNService"),
		0,
		NULL,
		REG_OPTION_NON_VOLATILE,
		KEY_ALL_ACCESS,
		NULL,
		&log_key,
		NULL);
	if (err != ERROR_SUCCESS) printf("Could register log message file.");

	err = RegSetValueEx(
		log_key,
		TEXT("EventMessageFile"),
		0,
		REG_SZ,
		(LPBYTE) szPath,
		(wcslen(szPath)+1)*sizeof(TCHAR));
	if (err != ERROR_SUCCESS) printf("Could not set message file.");

	err = RegCloseKey(log_key);
	if (err != ERROR_SUCCESS) printf("Could not close reg key.");

	CmdSetupService(argc, argv);
}

void CmdSetupService(int argc, char **argv)
{
	CHAR cmdLine[512];
	int i;
	HKEY key;
	LONG err;

	if (argc > 0)
	{
		cmdLine[0] = 0;
		for (i=0; i<argc; i++)
		{
			if (i == 0)
				strcat(cmdLine, "\"");
			else
				strcat(cmdLine, " \"");
			strcat(cmdLine, argv[i]);
			strcat(cmdLine, "\"");
		}

		err = RegCreateKeyA(HKEY_LOCAL_MACHINE, "SOFTWARE\\Clansoft\\SVNService", &key);
		if (err == ERROR_SUCCESS)
		{
			RegSetValueExA(key, "CommandLine", 0, REG_SZ, cmdLine, strlen(cmdLine)+1);
			RegCloseKey(key);
		}

		printf("Commandline set: %s\r\n", cmdLine);
	}
	else
	{
		printf("No parameters specified!\r\n");
	}
}


//
//  FUNCTION: CmdRemoveService()
//
//  PURPOSE: Stops and removes the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdRemoveService()
{
    SC_HANDLE   schService;
    SC_HANDLE   schSCManager;

    schSCManager = OpenSCManager(
                        NULL,                   // machine (NULL == local)
                        NULL,                   // database (NULL == default)
                        SC_MANAGER_ALL_ACCESS   // access required
                        );
    if ( schSCManager )
    {
        schService = OpenService(schSCManager, TEXT(SZSERVICENAME), SERVICE_ALL_ACCESS);

        if (schService)
        {
            // try to stop the service
            if ( ControlService( schService, SERVICE_CONTROL_STOP, &ssStatus ) )
            {
                _tprintf(TEXT("Stopping %s."), TEXT(SZSERVICEDISPLAYNAME));
                Sleep( 1000 );

                while( QueryServiceStatus( schService, &ssStatus ) )
                {
                    if ( ssStatus.dwCurrentState == SERVICE_STOP_PENDING )
                    {
                        _tprintf(TEXT("."));
                        Sleep( 1000 );
                    }
                    else
                        break;
                }

                if ( ssStatus.dwCurrentState == SERVICE_STOPPED )
                    _tprintf(TEXT("\n%s stopped.\n"), TEXT(SZSERVICEDISPLAYNAME) );
                else
                    _tprintf(TEXT("\n%s failed to stop.\n"), TEXT(SZSERVICEDISPLAYNAME) );

            }

            // now remove the service
            if( DeleteService(schService) )
                _tprintf(TEXT("%s removed.\n"), TEXT(SZSERVICEDISPLAYNAME) );
            else
                _tprintf(TEXT("DeleteService failed - %s\n"), GetLastErrorText(szErr,256));


            CloseServiceHandle(schService);
        }
        else
            _tprintf(TEXT("OpenService failed - %s\n"), GetLastErrorText(szErr,256));

        CloseServiceHandle(schSCManager);
    }
    else
        _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));
}




///////////////////////////////////////////////////////////////////
//
//  The following code is for running the service as a console app
//


//
//  FUNCTION: CmdDebugService(int argc, char ** argv)
//
//  PURPOSE: Runs the service as a console application
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdDebugService(int argc, char ** argv)
{
    DWORD dwArgc;
    LPTSTR *lpszArgv;

#ifdef UNICODE
    lpszArgv = CommandLineToArgvW(GetCommandLineW(), &(dwArgc) );
#else
    dwArgc   = (DWORD) argc;
    lpszArgv = argv;
#endif

    _tprintf(TEXT("Debugging %s.\n"), TEXT(SZSERVICEDISPLAYNAME));

    SetConsoleCtrlHandler( ControlHandler, TRUE );

    ServiceStart( dwArgc, lpszArgv );
}


//
//  FUNCTION: ControlHandler ( DWORD dwCtrlType )
//
//  PURPOSE: Handled console control events
//
//  PARAMETERS:
//    dwCtrlType - type of control event
//
//  RETURN VALUE:
//    True - handled
//    False - unhandled
//
//  COMMENTS:
//
BOOL WINAPI ControlHandler ( DWORD dwCtrlType )
{
    switch( dwCtrlType )
    {
        case CTRL_BREAK_EVENT:  // use Ctrl+C or Ctrl+Break to simulate
        case CTRL_C_EVENT:      // SERVICE_CONTROL_STOP in debug mode
            _tprintf(TEXT("Stopping %s.\n"), TEXT(SZSERVICEDISPLAYNAME));
            ServiceStop();
            return TRUE;
            break;

    }
    return FALSE;
}

//
//  FUNCTION: GetLastErrorText
//
//  PURPOSE: copies error message text to string
//
//  PARAMETERS:
//    lpszBuf - destination buffer
//    dwSize - size of buffer
//
//  RETURN VALUE:
//    destination buffer
//
//  COMMENTS:
//
LPTSTR GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize )
{
    DWORD dwRet;
    LPTSTR lpszTemp = NULL;

    dwRet = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_ARGUMENT_ARRAY,
                           NULL,
                           GetLastError(),
                           LANG_NEUTRAL,
                           (LPTSTR)&lpszTemp,
                           0,
                           NULL );

    // supplied buffer is not long enough
    if ( !dwRet || ( (long)dwSize < (long)dwRet+14 ) )
        lpszBuf[0] = TEXT('\0');
    else
    {
        lpszTemp[lstrlen(lpszTemp)-2] = TEXT('\0');  //remove cr and newline character
        _stprintf( lpszBuf, TEXT("%s (0x%x)"), lpszTemp, GetLastError() );
    }

    if ( lpszTemp )
        LocalFree((HLOCAL) lpszTemp );

    return lpszBuf;
}

volatile BOOL stop_service = FALSE;

VOID ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv)
{
	TCHAR szDllName[_MAX_PATH];
	TCHAR szApp[_MAX_PATH*2];
	TCHAR szCommand[_MAX_PATH*2];
	int i, len;
	STARTUPINFO startupInfo;
	PROCESS_INFORMATION processInfo;
	BOOL result;
	DWORD exitcode;
	LONG lResult;
	HKEY key;
	DWORD type, size;
	LPTSTR pos;

	ReportStatusToSCMgr(SERVICE_START_PENDING, NO_ERROR, 5000);
	stop_service = FALSE;

	GetModuleFileName(0, szDllName, _MAX_PATH);
	len = wcslen(szDllName);

	wcscpy(szApp, szDllName);
	for (i=len-1; i>0; i--)
	{
		if (szApp[i] == '\\')
		{
			szApp[i+1] = 0;
			break;
		}
	}
	_tprintf(TEXT("Setting working directory: %s\n"), szApp);
	SetCurrentDirectory(szApp);
	wcscat(szApp, L"svnserve.exe");

	wcscpy(szCommand, L"\"");
	wcscat(szCommand, szApp);
	wcscat(szCommand, L"\" ");

	lResult = RegOpenKey(HKEY_LOCAL_MACHINE, L"SOFTWARE\\Clansoft\\SVNService", &key);
	if (lResult == ERROR_SUCCESS)
	{
		pos = szCommand+wcslen(szCommand);
		size = (_MAX_PATH*2-wcslen(szCommand))*sizeof(TCHAR);
		type = REG_SZ;
		lResult = RegQueryValueEx(key, L"CommandLine", 0, &type, (LPBYTE) pos, &size);
		if (lResult != ERROR_SUCCESS || type != REG_SZ)
			wcscpy(pos, L"-d");
		else
			_tprintf(TEXT("RegQueryValueEx failed with error %d, type %d\n"), lResult, type);
		RegCloseKey(key);
	}
	else
	{
		_tprintf(TEXT("RegOpenKey failed with %d\n"), lResult);
		wcscat(szCommand, L"-d");
	}

	_tprintf(TEXT("Running command: %s\n"), szCommand);

	memset(&startupInfo, 0, sizeof(STARTUPINFO));
	startupInfo.cb = sizeof(STARTUPINFO);

	memset(&processInfo, 0, sizeof(PROCESS_INFORMATION));

	result = CreateProcess(
		szApp,
		szCommand,
		0,
		0,
		FALSE,
		CREATE_NEW_PROCESS_GROUP,
		0,
		0,
		&startupInfo,
		&processInfo);

	if (result == FALSE)
	{
		AddToMessageLog(Event_FailCreateProcess, szCommand, EVENTLOG_ERROR_TYPE);
		ReportStatusToSCMgr(SERVICE_STOPPED, GetLastError(), 1000);
		return;
	}

	if (processInfo.hThread)
		CloseHandle(processInfo.hThread);

	ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 1000);

	while (!stop_service)
	{
		if (WaitForSingleObject(processInfo.hProcess, 1000) != WAIT_TIMEOUT)
		{
			AddToMessageLog(Event_ProcessEnded, szCommand, EVENTLOG_ERROR_TYPE);
			break;
		}
	}

	if (processInfo.hProcess)
	{
		GenerateConsoleCtrlEvent(CTRL_C_EVENT, processInfo.dwProcessId);
		GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, processInfo.dwProcessId);
		ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 30000);
		if (WaitForSingleObject(processInfo.hProcess, 28000) == WAIT_TIMEOUT)
		{
			AddToMessageLog(Event_ProcessTimeout, szCommand, EVENTLOG_ERROR_TYPE);
			TerminateProcess(processInfo.hProcess, 255);
		}

		exitcode = STATUS_WAIT_0;
		GetExitCodeProcess(processInfo.hProcess, &exitcode);

		/* Unusual exit code?*/
/*		if (exitcode != STATUS_WAIT_0)
		{
			AddToMessageLog(Event_ExitCode, exitcode, EVENTLOG_WARNING_TYPE);
		}*/
		CloseHandle(processInfo.hProcess);
	}

	ReportStatusToSCMgr(SERVICE_STOPPED, NO_ERROR, 1000);
}

VOID ServiceStop()
{
	ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 2000);
	stop_service = TRUE;
}
